angular
  .module("gaugesScreen", [])

  .directive("bngMapRenderUncompressed", function () {
    return {
      template: `<svg width="100%" height="100%" class="container"></svg>`,
      scope: {
        map: "<",
        color: "@?",
        width: "@?",
        drivability: "@?",
      },
      replace: true,
      restrict: "E",
      link: function (scope, element, att8s) {
        "use strict";
        var svg = element[0],
          mapScale = 1,
          domElems = {},
          getColor = (rClass) =>
            scope.color || (rClass === 0 ? "black" : "white"); // if there is a color set use that otherwise use the defaults
        function isEmpty(obj) {
          return Object.keys(obj).length === 0;
        }

        function calcRadius(radius) {
          return Math.min(Math.max(radius, 0), 5) * 3;
        }

        scope.$watch("map", function (newVal) {
          if (newVal && !isEmpty(newVal)) {
            setupMap(newVal, angular.element(svg));
          }
        });

        function _createLine(p1, p2, color) {
          return hu("<line>", svg).attr({
            x1: p1.x,
            y1: p1.y,
            x2: p2.x,
            y2: p2.y,
            stroke: color,
            strokeWidth: Math.max(p1.radius, p2.radius),
            strokeLinecap: "round",
          });
        }

        function drawRoads(nodes, drivabilityMin, drivabilityMax) {
          var drawn = {};
          for (var key in nodes) {
            var el = nodes[key];
            // walk the links of the node
            if (el.links !== undefined) {
              // links
              for (var key2 in el.links) {
                var el2 = nodes[key2];
                var drivability = el.links[key2].drivability;
                if (el2 !== undefined) {
                  if (
                    drivability >= drivabilityMin &&
                    drivability <= drivabilityMax
                  ) {
                    drawn[
                      key + "." + key2 + drivabilityMin + drivabilityMax
                    ] = true;
                    if (
                      domElems[
                        key + "." + key2 + drivabilityMin + drivabilityMax
                      ] !== undefined
                    ) {
                      domElems[
                        key + "." + key2 + drivabilityMin + drivabilityMax
                      ].remove();
                    }
                    domElems[
                      key + "." + key2 + drivabilityMin + drivabilityMax
                    ] = _createLine(
                      {
                        x: el.pos[0] / mapScale,
                        y: -el.pos[1] / mapScale,
                        radius: calcRadius(el.radius),
                      },
                      {
                        x: el2.pos[0] / mapScale,
                        y: -el2.pos[1] / mapScale,
                        radius: calcRadius(el2.radius), // prevents massive blobs due to waypoints having larger radius'
                      },
                      getColor(drivability)
                    );
                  }
                }
              }
            }
          }

          // remove all elems that are from previous calls
          for (var key in domElems) {
            if (
              !drawn[key] &&
              key.endsWith("" + drivabilityMin + drivabilityMax)
            ) {
              domElems[key].remove();
              domElems[key] = undefined; // delete domNode reference and allow for gc
            }
          }
        }

        function setupMap(data) {
          if (data != null) {
            svg.setAttribute("viewBox", data.viewParams.join(" "));

            // use background image if existing, otherwise draw a simple grid
            if (data.minimapImage && data.terrainOffset && data.terrainSize) {
              // mapcontainer.style.backgroundSize = "100%"
              // mapcontainer.style.backgroundImage = "url('/" + data.minimapImage + "')"
              var bgImage = hu('<image>', svg).attr({
                'x': data.terrainOffset[0] / mapScale,
                'y': data.terrainOffset[1] / mapScale,
                'width': data.terrainSize[0] / mapScale,
                'height': data.terrainSize[1] / mapScale,
                // 'transform': "scale(-1,-1)",
                'xlink:href': "/" + data.minimapImage,
              }).prependTo(svg)

            } else {
              // draw grid
              var distX = maxX - minX
              var dx = 50
              for (var x = minX; x <= maxX + 1; x += dx) {
                _createLine({ x: x, y: minY, radius: 1 }, { x: x, y: maxY, radius: 1 }, '#FFFFFF55')
              }
              var distY = maxY - minY
              var dy = 50
              for (var y = minY; y <= maxY + 1; y += dy) {
                _createLine({ x: minX, y: y, radius: 1 }, { x: maxX, y: y, radius: 1 }, '#FFFFFF55')
              }
            }

            // draw dirt roads and then normal on top
            if (scope.drivability !== "false") {
              drawRoads(data.nodes, 0, 0.9);
              drawRoads(data.nodes, 0.9, 1);
            } else {
              drawRoads(data.nodes, 0, 1);
            }
          }
        }
      },
    };
  })

  .controller("GaugesScreenController", function ($scope, $element, $window) {
    "use strict";
    var vm = this;

    var svg;
    var navContainer = $element[0].children[0];
    var navDimensions = [];

    var speedoDisplay = {};
    var tacho = {};
    var gagues = {};
    var electrics = { lights: {}, dse: {} };
    var currentGear = "";

    var ready = false;

    var units = {
      uiUnitConsumptionRate: "metric",
      uiUnitDate: "ger",
      uiUnitEnergy: "metric",
      uiUnitLength: "metric",
      uiUnitPower: "hp",
      uiUnitPressure: "bar",
      uiUnitTemperature: "c",
      uiUnitTorque: "metric",
      uiUnitVolume: "l",
      uiUnitWeight: "kg",
    };
    var unitspeedConv = 3.6;

    function getNodesArray(prefix, number, root) {
      let arr = [];
      for (let i = 0; i < number; i++) {
        let node = hu(prefix + i, root);
        if (!node) console.error(prefix + i, "node not found");
        else arr.push(node);
      }
      return arr;
    }

    $scope.onSVGLoaded = function () {
      svg = $element[0].children[0].children[0];

      speedoDisplay.speedValue = hu("#speed_val", svg);
      speedoDisplay.speedUnit = hu("#speed_unit", svg);
      speedoDisplay.gears = hu("#gear_txt", svg);

      tacho.needle = hu("#needle", svg);

      electrics.lights.signal_L = hu("#light_signal_L", svg);
      electrics.lights.signal_R = hu("#light_signal_R", svg);
      electrics.lights_lights = hu("#light_lights", svg);
      electrics.lights_highbeam = hu("#light_highbeam", svg);
      electrics.lights.lowpressure = hu("#light_lowpressure", svg);
      electrics.lights.parkingbrake = hu("#light_parkingbrake", svg);
      electrics.lights.checkengine = hu("#light_checkengine", svg);
      electrics.lights.lowfuel = hu("#fuel_logo", svg);
      electrics.lights_battery = hu("#lights_battery", svg);
      electrics.odo_txt = hu("#odo_txt", svg);
      electrics.odo_txt1 = hu("#odo_txt1", svg);
      electrics.fuel_range = hu("#fuel_range", svg);
      electrics.envTime = hu("#env_time", svg);
      electrics.date = hu("#date", svg);
      electrics.envTemp = hu("#env_temp", svg);

      gagues.l_water = hu("#g_water", svg);
      gagues.l_turbo = hu("#g_turbo", svg);
      gagues.water = getNodesArray("#g_water", 8, gagues.l_water);
      gagues.turbo = getNodesArray("#g_turbo", 8, gagues.l_turbo);
      gagues.waterValue = hu("#water_val", svg);

      ready = true;
    };

    function limitVal(min, val, max) {
      return Math.min(Math.max(min, val), max);
    }
    function map_range(value, low1, high1, low2, high2) {
      return low2 + ((high2 - low2) * (value - low1)) / (high1 - low1);
    }
    const clamp = (num, min, max) => Math.min(Math.max(num, min), max);
    const clamp_remap = (val, min, max) =>
      map_range(clamp(val, min, max), min, max, 0, 1);

    function updateGearIndicator(data) {
      // only update when gear is changed
      if (currentGear !== data.electrics.gear) {
        currentGear = data.electrics.gear;
        if (isNaN(data.electrics.gear)) {
          //auto,DCT
          speedoDisplay.gears.text(data.electrics.gear);
        } else {
          //manuel
          switch (data.electrics.gear) {
            case -1:
              speedoDisplay.gears.text("R");
              break;
            case 0:
              speedoDisplay.gears.text("N");
              break;
            default:
              speedoDisplay.gears.text(data.electrics.gear);
              break;
          }
        }
      }
    }

    function updateTachoDisplays(data) {
      if (ready) {
        //106deg
        let angle =
          limitVal(-4000, data["electrics"]["rpmTacho"] - 4000, 4000) * 0.03;
        tacho.needle.n.style["transform"] = "rotate(" + angle + "deg)";
      }
    }

    function updateSpeedDisplays(data) {
      if (ready) {
        speedoDisplay.speedValue.text(
          (data.electrics.wheelspeed * unitspeedConv).toFixed(0)
        );
      }
    }

    function updateGagues(data) {
      if (ready) {
        let remaped = clamp_remap(data.electrics.watertemp, 60, 120);
        let tmp = UiUnits.temperature(data.electrics.watertemp);

        gagues.waterValue.text(tmp.val.toFixed(0) + tmp.unit);

        for (let i = 0; i < gagues.water.length; i++) {
          let threadshold = (i + 0.5) / gagues.water.length;
          if (i >= 5)
            gagues.water[i].css({
              fill: remaped > threadshold ? "#f00" : "#323232",
            });
          else
            gagues.water[i].css({
              fill: remaped > threadshold ? "#fff" : "#323232",
            });
        }

        remaped = clamp_remap(data.electrics.turboBoost, 0, 30);

        for (let i = 0; i < gagues.turbo.length; i++) {
          let threadshold = (i + 0.5) / gagues.turbo.length;
          if (i >= 6)
            gagues.turbo[i].css({
              fill: remaped > threadshold ? "#f00" : "#323232",
            });
          else
            gagues.turbo[i].css({
              fill: remaped > threadshold ? "#fff" : "#323232",
            });
        }
      }
    }

    // overwriting plain javascript function so we can access from within the controller

    $window.setup = (data) => {
      if (!ready) {
        console.log("calling setup while svg not fully loaded");
        setTimeout(function () {
          $window.setup(data);
        }, 100);
        return;
      }

      //console.log("setup",data);
      for (let dk in data) {
        if (typeof dk == "string" && dk.startsWith("uiUnit")) {
          units[dk] = data[dk];
        }
      }
      vueEventBus.emit("SettingsChanged", { values: units });

      if (units.uiUnitLength == "metric") {
        speedoDisplay.speedUnit.text("KMH");
        unitspeedConv = 3.6;
      } else {
        speedoDisplay.speedUnit.text("MPH");
        unitspeedConv = 2.23694;
      }
    };

    $window.initMap = (data) => {
      navDimensions = data.viewParams = [
        data.terrainOffset[0],
        data.terrainOffset[1],
        data.terrainSize[0],
        data.terrainSize[1],
      ];

      
      $scope.$apply(() => {
        vm.mapData = data;
      });

      navContainer.style.width = data.terrainSize[0] + "px";
      navContainer.style.height = data.terrainSize[1] + "px";
    };

    $window.updateMap = (data) => {
      var focusX = -data.x;
      var focusY = data.y;
      var origin = `${navDimensions[0] * -1 - focusX}px ${
        navDimensions[1] * -1 - focusY
      }px`;
      navContainer.style.transformOrigin = origin;
      var translateX = navDimensions[0] + 512 + focusX;
      var translateY = navDimensions[1] + 256 + focusY;
      navContainer.style.transform = `translate3d(${translateX}px,${translateY}px, 0px) rotateX(${55}deg) rotateZ(${
        180 + (data.rotation + 360)
      }deg) scale(1)`;
    };

    function setElec(node, state, key) {
      if (node === undefined || node === null) {
        console.error("setElec: svg element not found", key);
        return;
      }
      if (state === undefined || state === null) {
        console.error("setElec: state not found", key);
        node.n.style.display = "none";
        return;
      }
      node.n.style.filter = state === true || state > 0.1 ? "" : "url(#off)";
      //node.n.setAttribute("opacity", (state || state>0.1)?1.0:0.3)
    }

    $window.updateElectrics = (data) => {
      for (var k in electrics.lights) {
        setElec(electrics.lights[k], data.electrics[k], k);
      }

      electrics.lights_lights.n.style.display =
        data.electrics.lights > 0.5 && data.electrics.lights < 1.5
          ? "inline"
          : "none";
      electrics.lights_highbeam.n.style.display =
        data.electrics.lights > 0.5 && data.electrics.lights < 1.5
          ? "none"
          : "inline";
      electrics.lights_highbeam.n.style.filter =
        data.electrics.lights > 1 ? "" : "url(#off)";

      electrics.lights_battery.n.style.filter =
        data.electrics.engineRunning < 0.1 ? "" : "url(#off)";

      let tmp = UiUnits.distance(
        data.customModules.combustionEngineData.remainingRange * 1000
      );
      if (data.customModules.combustionEngineData.remainingRange < 5)
        electrics.fuel_range.text("---");
      else electrics.fuel_range.text(tmp.val.toFixed(0) + tmp.unit);

      if (data.electrics.odometer) {
        let val = data.electrics.odometer;
        val *= units.uiUnitLength == "metric" ? 0.001 : 0.0006215;
        electrics.odo_txt.text(
          val.toFixed(1) + " " + (units.uiUnitLength == "metric" ? "km" : "mi")
        );
        val = val + 3981.3;
        electrics.odo_txt1.text(
          val.toFixed(1) + " " + (units.uiUnitLength == "metric" ? "km" : "mi")
        );
      }
      electrics.envTime.text(data.customModules.environmentData.time);
      electrics.date.text(data.customModules.environmentData.time);
      let envTempC = Number(
        data.customModules.environmentData.temperatureEnv
      );
      let envTempF =
        Number(data.customModules.environmentData.temperatureEnv) * 1.8 + 32;
      let envTemp = units.uiUnitLength == "metric" ? envTempC : envTempF;
      electrics.envTemp.text(
        envTemp.toFixed(0) + UiUnits.temperature(0).unit
      );
    };

    $window.updateMode = (data) => {
      if(!ready){
        console.log("calling updateMode while svg not fully loaded");
        setTimeout(function(){ $window.updateMode(data) }, 100);
        return;
      }
    }

    $window.updateData = (data) => {
      if (data) {
        if (!ready) {
          console.log("not ready");
          return;
        }
        // Update PRNDS display
        updateGearIndicator(data);
        // Update Speed displays
        updateSpeedDisplays(data);
        updateTachoDisplays(data);
        updateElectrics(data);
        updateGagues(data);
      }
    };
  });
